from __future__ import unicode_literals
import json
from sqlalchemy import asc, desc, text, types
from sqlalchemy.exc import OperationalError, InvalidRequestError
from sqlalchemy.sql import expression, functions


def recursive_find(data, key):
    result = []
    if isinstance(data, dict):
        for k, v in data.items():
            if k == key:
                result.append(v)
            else:
                result += recursive_find(v, key)
    elif isinstance(data, list):
        for item in data:
            result += recursive_find(item, key)
    return result


class Filter(object):
    def __init__(self, database, columns=None):
        self.database = database
        self.queries = None
        self.selected_columns = columns
        self._query()

    def distinct(self, distinct=True):
        for query in self.queries:
            query['rows'] = query['rows'].distinct()
        return self

    def where_sql(self, sql):
        return self.where_sqls([sql])

    def where_sqls(self, sqls):
        # sqls is list of sql text 'where'-style conditions

        active_queries = []
        for query in self.queries:
            for sql in sqls:
                query['rows'] = query['rows'].filter(text(sql))
            try:
                query['rows'].limit(1).all()  # inefficient
                active_queries.append(query)
            except OperationalError:
                # should cache these and show if no results at all found
                continue
        self.queries = active_queries
        return self

    def where_kv(self, conditions):
        # conditions is a dict
        active_queries = []
        for query in self.queries:
            try:
                query['rows'] = query['rows'].filter_by(**conditions)
                active_queries.append(query)
            except InvalidRequestError:
                continue
        self.queries = active_queries
        return self

    def select_from(self, sql_text):
        rows = self.database.session.execute(text(sql_text))
        self.queries = [
            {
                'table_name': 'custom',
                'table': rows._metadata,
                'rows': rows
            }
        ]
        return self

    def grep(self, pattern, case_sensitive=False):
        for query in self.queries:
            query['rows'] = self._add_grep(query['table'], query['rows'], pattern,
                                           case_sensitive=case_sensitive)
        return self

    def limit(self, limit):
        for query in self.queries:
            query['rows'] = query['rows'].limit(limit)
        return self

    def _query(self):

        table_items = self.database.tables_metadata.items()

        self.queries = []

        for table_name, table in sorted(table_items):

            if self.database.tables is not None:
                if table_name not in self.database.tables:
                    continue

            if self.selected_columns is not None:
                ok = True
                for name in self.selected_columns:
                    if name not in table.c:
                        ok = False
                if not ok:
                    continue
                rows = self.database.session.query(*[table.c[name]
                                                     for name in self.selected_columns])
            else:
                rows = self.database.session.query(table)
            self.queries.append({
                'table_name': table_name,
                'table': table,
                'rows': rows
            })

    def _add_grep(self, table, query, sequence, case_sensitive):
        # functions.concat would be neater, but doesn't seem to translate
        # correctly on sqlite
        parts = ''
        for idx, column in enumerate(table.columns):
            if not self.ok_column(column.name):
                continue
            if parts != '':
                parts = parts + ' // '
            part = functions.coalesce(expression.cast(column,
                                        types.Unicode),
                                        '')
            parts = parts + part
        if case_sensitive:
            query = query.filter(parts.contains(sequence))
        else:
            query = query.filter(parts.ilike('%%' + sequence + '%%'))
        return query

    def _default_order(self):
        for query in self.queries:
            table = query['table']
            primary_key = table.primary_key
            if self.selected_columns:
                query['rows'] = query['rows'].order_by(*[table.c[name]
                                                         for name in self.selected_columns])
            elif len(primary_key) >= 1:
                query['rows'] = query['rows'].order_by(*primary_key)
            elif len(table.c) >= 1:
                query['rows'] = query['rows'].order_by(*table.c)
        return self

    def order(self, ordering=None):
        if ordering is None:
            return self._default_order()
        for query in self.queries:
            table = query['table']
            orders = []
            for name in ordering:
                c = name[-1]
                order = name
                if c == '+' or c == '-':
                    order = order[:-1]
                order = table.c[order]
                if c == '+':
                    order = asc(order)
                elif c == '-':
                    order = desc(order)
                orders.append(order)
            if len(orders) > 0:
                query['rows'] = query['rows'].order_by(*orders)
        return self

    def where_kv_with_expansion(self, conditions):
        # conditions is a dict
        active_queries = []
        for query in self.queries:
            try:
                table = query['table']
                for key, val in conditions.items():
                    if key not in table.c:
                        raise InvalidRequestError('key not present')
                    if val == "" or val[0] != '@':
                        query['rows'] = query['rows'].filter(table.c[key] == val)
                    else:
                        file_filter = val[1:]
                        with open(file_filter, 'r') as fin:
                            data = json.load(fin)
                            vals = recursive_find(data, key)
                            query['rows'] = query['rows'].filter(table.c[key].in_(vals))
                active_queries.append(query)
            except InvalidRequestError:
                continue
        self.queries = active_queries
        return self

    def __iter__(self):
        return self.queries.__iter__()

    def __len__(self):
        return len(self.queries)

    def __getitem__(self, idx):
        return self.queries[idx]

    def ok_column(self, name):
        if self.selected_columns:
            if name not in self.selected_columns:
                return False
        return True

    @property
    def rows(self):
        if len(self) > 1:
            raise KeyError("more than one table in results")
        if len(self) == 0:
            return []
        return self[0]['rows']

    @property
    def row(self):
        rows = self.rows.all()
        if len(rows) > 1:
            raise KeyError("more than one row in results")
        return rows[0]



import re
from six import string_types


class Nullify(object):

    def __init__(self):
        self.need_underscore = re.compile(r'^_*NULL$')
        self.has_underscore = re.compile(r'^_(_*NULL)$')
        pass

    def stringy(self, value):
        return isinstance(value, string_types)

    def encode_null(self, value):
        if value is None:
            return 'NULL'
        if not self.stringy(value):
            return value
        if self.need_underscore.match(value):
            return '_{}'.format(value)
        return value

    def decode_null(self, value):
        if value is None:
            return value
        if not self.stringy(value):
            return value
        if value == 'NULL':
            return None
        result = self.has_underscore.match(value)
        if result:
            return result.group(1)
        return value



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

import argparse
import errno
from collections import OrderedDict
from datetime import datetime
from io import StringIO, BytesIO
import json
import openpyxl
import openpyxl.styles
import os
from shutil import copyfile
from sqlalchemy import Column, MetaData, Table, types
from sqlalchemy.exc import (CompileError, SAWarning)
from subprocess import call
import sys
import warnings

from catsql import cmdline
from catsql.database import Database
from catsql.nullify import Nullify
from catsql.patch import patchsql

if sys.version_info[0] == 2:
    import unicodecsv as csv
else:
    import csv

warnings.simplefilter("ignore", category=SAWarning)


# Get approximate length of header
class CsvRowWriter(object):
    def __init__(self):
        self.writer = None

    def writerow(self, row):
        row = [str(s) for s in row]
        try:
            queue = StringIO()
            self.writer = csv.writer(queue, lineterminator='')
            self.writer.writerow(row)
        except TypeError:
            queue = BytesIO()
            self.writer = csv.writer(queue, lineterminator='')
            self.writer.writerow(row)
        return queue.getvalue()


class SmartFormatter(argparse.HelpFormatter):

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        return argparse.HelpFormatter._split_lines(self, text, width)


class Viewer(object):
    def __init__(self, args, remainder, sys_args):
        self.args = args
        self.sys_args = sys_args
        self.failure = False
        self.values = None
        self.setup_filters(args)
        self.connect_database()
        self.process_remainder(remainder)

    def decomma(self, lst):
        if not lst:
            return lst
        lst = [l.split(',') for l in lst]
        return [item for sublist in lst for item in sublist]

    def setup_filters(self, args):
        self.url = args.catsql_database_url
        self.tables = self.decomma(args.table)
        self.selected_columns = self.decomma(args.column)
        self.ordering = self.decomma(args.order)

        self.context_columns = set()
        self.context_filters = dict()
        if args.load_bookmark:
            with open(self.url, 'r') as fin:
                nargs = json.loads(fin.read())
            self.url = nargs['url']
            self.tables = set(nargs['table'])
            self.args.distinct = nargs.get('distinct', False)
            self.selected_columns = nargs.get('column', [])
            self.context_filters = nargs['context']
            self.context_columns = set(nargs['hidden_columns'])
            if args.value is None:
                self.args.value = []

        if self.tables is not None:
            self.tables = set(self.tables)
        self.row_filter = args.sql
        self.output_in_csv = args.csv
        self.output_in_json = args.json
        self.output_in_sqlite = args.sqlite
        self.output_in_excel = args.excel

        self.target_db = None
        if self.output_in_sqlite:
            self.target_db = Database(self.output_in_sqlite[0], can_create=True)
        self.target_ss = None
        if self.output_in_excel:
            self.target_ss = openpyxl.Workbook()
            self.target_ss.remove_sheet(self.target_ss.active)

        if args.value is not None:
            for context in args.value:
                if '=' in context:
                    key, value = context.split('=', 1)
                    self.context_filters[key] = value
                    self.context_columns.add(key)
                else:
                    self.context_columns.add(context)

    def connect_database(self):
        database = Database(self.url, verbose=self.args.verbose, tables=self.tables)
        self.database = database
        self.url = self.args.catsql_database_url = database.full_url

    def process_remainder(self, remainder):
        column_names = set()
        if len(remainder) == 0:
            return
        for table in self.database.tables_metadata.values():
            column_names |= set(table.columns.keys())
        if 'catsql_database_url' in column_names:
            column_names.remove('catsql_database_url')
        parser = argparse.ArgumentParser()
        cmdline.add_options(parser)
        for name in sorted(column_names):
            try:
                parser.add_argument('--{}'.format(name), nargs=1, default=None)
            except argparse.ArgumentError:
                column_names.remove(name)
        self.values = vars(parser.parse_args(self.sys_args))
        dud = []
        for key, val in self.values.items():
            if (key not in column_names) or val is None:
                dud.append(key)
        for key in dud:
            del self.values[key]
        for key, val in self.values.items():
            self.values[key] = val[0]
            self.context_columns.add(key)

    def ok_column(self, name):
        if self.args.terse:
            if name in self.context_columns:
                return False
        if self.selected_columns:
            if name not in self.selected_columns:
                return False
        return True

    def start_table(self, table_name, columns):
        self.header_shown = False
        self.header_considered = False
        self.table_name = table_name
        self.columns = columns
        if self.selected_columns:
            self.columns = self.selected_columns

    def show_header_on_need(self):
        if self.header_shown or self.header_considered:
            return True
        self.header_considered = True
        if len(self.tables_so_far) > 0:
            if (self.output_in_csv or self.output_in_json or self.output_in_sqlite or
                  self.output_in_excel):
                if not self.output_in_sqlite:
                    self.failure = True
                self.tables_so_far.append(self.table_name)
                return False
            print("", file=self.output_file)
        if not (self.output_in_csv or self.output_in_json or self.output_in_sqlite or
                self.output_in_excel):
            print('== {} =='.format(self.table_name), file=self.output_file)
        if not (self.output_in_json or self.output_in_sqlite or self.output_in_excel):
            header_writer = CsvRowWriter()

            header = header_writer.writerow(list(column for column in self.columns
                                                 if self.ok_column(column)))
            print(header, file=self.output_file)
            if not self.output_in_csv:
                print('-' * len(header), file=self.output_file)
        self.header_shown = True
        self.tables_so_far.append(self.table_name)
        return True

    def show(self):

        self.tables_so_far = []

        work = None
        output_filename = None
        self.output_file = sys.stdout
        work_file = None
        self.start_table(None, None)

        try:
            if self.args.edit:
                import tempfile
                work = tempfile.mkdtemp()
                output_filename = os.path.join(work, 'reference.csv')
                work_file = open(output_filename, 'wt')
                self.output_file = work_file
                self.output_in_csv = True
                self.args.safe_null = True
                self.args.save_bookmark = [os.path.join(work, 'bookmark.json')]
            elif self.args.output:
                self.output_file = open(self.args.output[0], 'wt')

            viable_tables = []

            q = self.database.query(columns=self.selected_columns)
            touchable = True
            if self.args.select_from:
                q = q.select_from(self.args.select_from[0])
                touchable = False
            if self.args.distinct:
                q = q.distinct()
            if self.row_filter is not None:
                q = q.where_sqls(self.row_filter)
            if self.args.value is not None:
                q = q.where_kv(self.context_filters)
            if self.values is not None:
                q = q.where_kv_with_expansion(self.values)
            if self.args.grep:
                for pattern in self.args.grep:
                    q = q.grep(pattern, case_sensitive=False)
            if self.args.order:
                if 'none' not in self.args.order:
                    q.order(self.ordering)
            elif touchable:
                q.order()
            if self.args.limit:
                q = q.limit(int(self.args.limit[0]))

            ts = list(q)
            ts.sort(key=lambda t: t['table_name'])

            for t in ts:
                table_name = t['table_name']
                table = t['table']
                rows = t['rows']

                self.header_shown = False
                self.start_table(table_name, table.columns.keys())
                viable_tables.append(table_name)

                if self.args.types:
                    column_types = []
                    for name in self.columns:
                        if not self.ok_column(name):
                            continue
                        try:
                            column = table.c[name]
                            sql_name = str(column.type)  # make sure not nulltype
                        except CompileError:
                            sql_name = None
                        column_types.append(sql_name)
                    rows = [column_types]

                if self.target_ss:
                    ws = self.target_ss.create_sheet()
                    ws.title = table_name
                    ws.append(column for column in self.columns
                                if self.ok_column(column))
                    for row in rows:
                        ws.append(cell for c, cell in enumerate(row)
                                if self.ok_column(self.columns[c]))
                    s = openpyxl.styles.NamedStyle(name="Header",
                                                   font=openpyxl.styles.Font(bold=True))
                    for cell in next(ws.rows):
                        cell.style = s
                    for column_cells in ws.columns:
                        ws.column_dimensions[column_cells[0].column].auto_size = True

                if self.target_db:
                    if table_name in self.target_db.tables_metadata.keys():
                        # clear previous results
                        self.target_db.tables_metadata[table_name].drop(self.target_db.engine)
                    target = {'table': None, 'rows': []}

                    def fallback_type(example):
                        if isinstance(example, bool):
                            return types.Boolean
                        elif isinstance(example, int):
                            return types.Integer
                        elif isinstance(example, float):
                            return types.Float
                        elif isinstance(example, datetime):
                            return types.DateTime
                        return types.UnicodeText

                    def create_table(data):
                        if target['table'] is not None:
                            return
                        columns = []
                        for name in self.columns:
                            if not self.ok_column(name):
                                continue
                            column = table.c[name]
                            sql_type = column.type
                            try:
                                self.target_db.engine.dialect.type_compiler.process(sql_type)
                            except CompileError:
                                # some types need to be approximated
                                sql_type = None
                            if sql_type is None or isinstance(sql_type, types.NullType):
                                example = data.get(name)
                                sql_type = fallback_type(example)
                            columns.append(Column(name, sql_type,
                                                  primary_key=column.primary_key))
                        metadata = MetaData(bind=self.target_db.engine)
                        target['table'] = Table(table_name, metadata, *columns)
                        target['table'].create(self.target_db.engine)

                    def add_row(data):
                        if data:
                            target['rows'].append(data)
                        if len(target['rows']) > 10000 or not data:
                            target['table'].insert().execute(target['rows'])
                            target['rows'] = []

                    def sqlited(data):
                        if isinstance(data, dict) or isinstance(data, list):
                            return json.dumps(data)
                        return data

                    for row in rows:
                        data = dict((self.columns[c], sqlited(cell))
                                    for c, cell in enumerate(row)
                                    if self.ok_column(self.columns[c]))
                        create_table(data)
                        add_row(data)
                    create_table({})
                    add_row(None)

                if self.output_in_json or self.output_in_sqlite or self.output_in_excel:
                    if not self.show_header_on_need():
                        continue
                    if self.output_in_json:
                        self.save_as_json(table, rows, self.output_in_json[0])
                elif not self.args.count:
                    # csv spec is that eol is \r\n; we ignore this for our purposes
                    # for good reasons that unfortunately there isn't space to describe
                    # here on the back of this envelope
                    csv_writer = csv.writer(self.output_file, lineterminator='\n')
                    if self.args.safe_null:
                        nullify = Nullify()
                        for row in rows:
                            if not self.show_header_on_need():
                                continue
                            csv_writer.writerow(list(nullify.encode_null(cell)
                                                     for c, cell in enumerate(row)
                                                     if self.ok_column(self.columns[c])))
                    else:
                        for row in rows:
                            if not self.show_header_on_need():
                                continue
                            csv_writer.writerow(list(cell for c, cell in enumerate(row)
                                                     if self.ok_column(self.columns[c])))
                    del csv_writer
                else:
                    self.show_header_on_need()
                    ct = rows.count()
                    print("({} row{})".format(ct, '' if ct == 1 else 's'), file=self.output_file)

            if len(self.tables_so_far) == 0 and len(viable_tables) == 1:
                self.show_header_on_need()

            if self.args.save_bookmark:
                with open(self.args.save_bookmark[0], 'w') as fout:
                    link = OrderedDict()
                    link['url'] = self.args.catsql_database_url
                    link['table'] = list(self.tables) if self.tables else None
                    link['column'] = self.selected_columns
                    link['distinct'] = self.args.distinct
                    link['context'] = self.context_filters
                    link['hidden_columns'] = sorted(self.context_columns)
                    link['sql'] = self.args.sql
                    fout.write(json.dumps(link, indent=2))

            if self.args.edit and not self.failure:
                work_file.close()
                work_file = None
                edit_filename = os.path.join(work, 'variant.csv')
                copyfile(output_filename, edit_filename)
                editor = os.environ.get('TABLE_EDITOR', None)
                if not editor:
                    editor = os.environ.get('EDITOR', 'nano')
                call([editor, edit_filename])
                patchsql([self.url, '--table'] + self.tables_so_far +
                         ['--follow', output_filename, edit_filename,
                            '--safe-null'] +
                            (['--quiet'] if self.args.quiet else []))

        finally:
            if self.failure:
                print("ERROR: "
                      "More than one table in csv/json output, consider adding:",
                      file=sys.stderr)
                for name in self.tables_so_far:
                    print("  --table {}".format(name),
                          file=sys.stderr)

            if self.target_ss:
                self.target_ss.save(self.output_in_excel[0])
            if work:
                if work_file:
                    try:
                        work_file.close()
                    except:
                        pass
                    work_file = None
                import shutil
                shutil.rmtree(work)
                work = None

    def save_as_json(self, table, rows, filename):
        result = OrderedDict()
        result['count'] = 0
        result['meta'] = {
            'generator': 'catsql'
        }
        results = result['results'] = []
        names = []
        idxs = []
        for idx, column in enumerate(self.columns):
            if not self.ok_column(column):
                continue
            idxs.append(idx)
            names.append(column)
        for row in rows:
            od = OrderedDict()
            for i, idx in enumerate(idxs):
                od[names[i]] = row[idx]
            results.append(od)
        result['count'] = len(results)
        with open(filename, 'w') as fout:
            fout.write(json.dumps(result, indent=2))


def catsql(sys_args):

    parser = argparse.ArgumentParser(description='Quickly display and edit a slice of a database.',
                                     formatter_class=SmartFormatter)

    cmdline.add_options(parser)
    args, remainder = parser.parse_known_args(sys_args)
    viewer = Viewer(args, remainder, sys_args)
    viewer.show()


def main():
    try:
        catsql(sys.argv[1:])
    except IOError as error:
        if error.errno == errno.EPIPE:
            # totally benign e.g. pipe through head/tail
            pass
        else:
            raise


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import print_function
import argparse
import daff
import json
from sqlalchemy.exc import SAWarning
import sys
import warnings

from catsql.daffsql.sqlalchemy_database import SqlAlchemyDatabase
from catsql.nullify import Nullify

if sys.version_info[0] == 2:
    import unicodecsv as csv
else:
    import csv

warnings.simplefilter("ignore", category=SAWarning)


def fix_nulls(table, active):
    if not active:
        return
    nullify = Nullify()
    for row in table:
        for i in range(0, len(row)):
            row[i] = nullify.decode_null(row[i])


def patchsql(sys_args):

    parser = argparse.ArgumentParser(description='Patch a database.')

    parser.add_argument('url', help='Sqlalchemy-compatible database url')

    parser.add_argument('--patch', nargs=1, required=False, default=None,
                        help="A csv file describing the patch. In the "
                        "format output by daff.")

    parser.add_argument('--follow', nargs=2, required=False, default=None,
                        help="An alternative to --patch option.  Specify"
                        "two csv files to compare, and patch from their diff.")

    parser.add_argument('--table', nargs=1, required=True, default=None,
                        help='Table to which patch should be applied.')

    parser.add_argument('--safe-null', required=False, action='store_true',
                        help='Decode nulls in a reversible way.')

    parser.add_argument('--quiet', required=False, action='store_true',
                       help='Do not show computed diff.')

    args = parser.parse_args(sys_args)

    url = args.url
    tables = args.table

    db = SqlAlchemyDatabase(url)
    st = daff.SqlTable(db, daff.SqlTableName(tables[0]))

    patch = None

    if args.patch:
        with open(args.patch[0], 'rt') as fin:
            patch = list(csv.reader(fin))
            patch = daff.Coopy.tablify(patch)

    if args.follow:
        with open(args.follow[0], 'rt') as fin:
            table0 = list(csv.reader(fin))
            fix_nulls(table0, args.safe_null)
        with open(args.follow[1], 'rt') as fin:
            table1 = list(csv.reader(fin))
            fix_nulls(table1, args.safe_null)
        patch = daff.Coopy.diff(table0, table1)
        ansi_patch = daff.Coopy.diffAsAnsi(table0, table1)
        if not args.quiet:
            print(ansi_patch, file=sys.stderr, end='')

    if not patch:
        raise KeyError('please specify either --patch or --follow')

    daff_patch = daff.HighlightPatch(st, patch)
    daff_patch.apply()
    if db.events['skips'] != 0:
        print(" * {}".format(json.dumps(db.events),
                             file=sys.stderr))


def main():
    patchsql(sys.argv[1:])


if __name__ == "__main__":
    main()



from catsql.filter import Filter
import os
from sqlalchemy import create_engine
from sqlalchemy.exc import ArgumentError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import create_session


class Database(object):
    def __init__(self, url, verbose=False, tables=None, can_create=False):
        self.url = url
        self._full_url = self.url
        self.verbose = verbose
        self.tables = tables
        self.can_create = can_create
        self.connect_database()

    def connect_database(self):
        self.Base = declarative_base()
        try:
            self.engine = create_engine(self.url, echo=self.verbose)
        except ImportError as e:
            print("Support library for this database not installed - {}".format(e))
            exit(1)
        except ArgumentError:
            if os.path.exists(self.url) or self.can_create:
                try:
                    # maybe this is a local sqlite database?
                    sqlite_url = 'sqlite:///{}'.format(self.url)
                    self.engine = create_engine(sqlite_url, echo=self.verbose)
                    self._full_url = sqlite_url
                except ArgumentError:
                    # no joy, recreate the original problem and die.
                    self.engine = create_engine(self.url, echo=self.verbose)
            else:
                raise
        only = None
        if self.tables:
            only = list(self.tables)
        self.Base.metadata.reflect(self.engine, only=only)

        self._session = create_session(bind=self.engine)

    @property
    def session(self):
        return self._session

    @property
    def full_url(self):
        return self._full_url

    @property
    def tables_metadata(self):
        return self.Base.metadata.tables

    def query(self, columns=None):
        return Filter(self, columns=columns)
